(* File: mirror.mli

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

type mirror = float * float * float
(** components of a plane equation passing through zero (ax + by + cx +
    d = 0, we know d = 0 since a * 0 + b * 0 + c * 0 + d = 0, so we only
    need to store a, b and c)

    We know there is only a limited number of mirrors for each symmetry
    group, so using them is a good way to avoid numerical error since we
    will always use the same mirrors (the first calculated, which are the
    more accurate)

    An additionnal invariant is that the norm of the vector (a, b, c) must
    be equal to one. This simplifies calculus in the development. Use the
    mirror making functions to ensure having a correct mirror
*)

val make_normal : float -> float -> float -> mirror
(** Constructs a mirror from a normal vector *)

val make_points : Point.point -> Point.point -> mirror
(** Constructs a mirror from two distinct points other than the
    origin *)

val eval : mirror -> Point.point -> float
(** Evaluates the point on the equation of the mirror *)

val distance : mirror -> Point.point -> float
(** Computes the distance from a point to the mirror *)

val image : mirror -> Point.point -> Point.point
(** [image m p] Returns the image of [p] mirrored into [m] *)
