(* File: operations.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

module P = Point
module W = Wythoff

let faces_at_vertex p i =
  List.filter (List.exists (( = ) i)) p.W.faces

let faces_at_edge p (i1, i2) =
  List.filter (List.exists (( = ) i2))
    (List.filter (List.exists (( = ) i1)) p.W.faces)

(* Important: ORDERED edges ! Two edges are adjacent in the result if and
   only if they are issued from a common vertex, and the first and last
   edges of the result are issued from a common vertex *)
let edges_of_face _ face =
  let rec aux = function
    | [] -> []
    | hd :: [] -> [List.hd face, hd]
    | hd1 :: hd2 :: tl -> (min hd1 hd2, max hd1 hd2) :: (aux (hd2 :: tl))
  in
  aux face

let adjacent_faces p face =
  let edges = edges_of_face p face in
  let faces = List.flatten (List.map (faces_at_edge p) edges) in
  List.filter (( <> ) face) faces

let edges_at_vertex p i =
  List.filter (fun (a, b) -> a = i || b = i) p.W.edges

let adjacent_vertices p i =
  let edges = edges_at_vertex p i in
  List.map (fun (a, b) -> if a = i then b else a) edges

type operation =
  | Copy
  | Rectify
  | Dualize
  | Truncate of float
  | Bitruncate
  | Cantellate of float
  | Omnitruncate
  | Snub
  | Snub_right

let copy {W.vertices=vertices; W.edges=edges; W.faces=faces} =
  { W.vertices = Array.init (Array.length vertices) (fun i -> vertices.(i));
    W.edges = List.filter (fun _ -> true) edges;
    W.faces = List.map (fun f -> List.filter (fun _ -> true) f) faces }

let rectify p =
  (* New vertices indices are the indices of actual edges *)
  let new_vertices = Array.map
    (fun (a, b) -> P.center [| p.W.vertices.(a); p.W.vertices.(b) |])
    (Array.of_list p.W.edges)
  in
  let rec edge_index ?(i=0) edges edge = match edges with
    | [] -> assert false
    | hd :: _ when hd = edge -> i
    | hd :: tl  -> edge_index ~i:(succ i) tl edge
  in
  let faces_vertices = List.map
    (fun v -> List.map (edge_index p.W.edges) (edges_at_vertex p v))
    (Array.to_list (Array.mapi (fun i _ -> i) p.W.vertices))
  in
  let faces_transformed = List.map
    (fun f -> List.map (edge_index p.W.edges) (edges_of_face p f))
    p.W.faces
  in
  { W.vertices = new_vertices;
    W.edges = List.flatten (List.map (edges_of_face p) faces_transformed);
    W.faces = faces_vertices @ faces_transformed }

let dualize p = failwith "Not yet implemented"

let truncate ?(dist=0.5) p = failwith "Not yet implemented"

let bitruncate p = failwith "Not yet implemented"

let cantellate ?(dist=0.5) p = failwith "Not yet implemented"

let omnitruncate p = failwith "Not yet implemented"

let snub p = failwith "Not yet implemented"

let snub_right p = failwith "Not yet implemented"

let apply = function
  | Copy -> copy
  | Rectify -> rectify
  | Dualize -> dualize
  | Truncate dist -> truncate ~dist
  | Bitruncate -> bitruncate
  | Cantellate dist -> cantellate ~dist
  | Omnitruncate -> omnitruncate
  | Snub -> snub
  | Snub_right -> snub_right
