(* File: point.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

open Common

type point =
  | Spherical of (float * float)
  | Cartesian of (float * float * float)

let to_cartesian = function
  | Spherical (phi, theta) ->
    Cartesian (sin phi *. cos theta, sin phi *. sin theta, cos phi)
  | p -> p

let to_spherical = function
  | Cartesian (x, y, z) -> Spherical (acos z, atan2 y x)
  | p -> p

let rec ( =* ) p p' = match (p, p') with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    x =~ x' && y =~ y' && z =~ z'
  | Spherical (a, b), Spherical (a', b') ->
    a =~ a' && b =~ b'
  | p, p' -> to_cartesian p =* to_cartesian p'

let rec distance p p' = match (p, p') with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    let dx = x -. x' and dy = y -. y' and dz = z -. z' in
    sqrt (dx *. dx +. dy *. dy +. dz *. dz)
  | p, p' -> distance (to_cartesian p) (to_cartesian p')

let rec to_string ?(after="") = function
  | Cartesian (x, y, z) -> Printf.sprintf "(%.10f, %.10f, %.10f)%s" x y z after
  | (Spherical _) as p -> to_string (to_cartesian p)

let center points =
  let sx = ref 0. and sy = ref 0. and sz = ref 0. in
  let factor = 1. /. float (Array.length points) in
  Array.iter (function
  | Cartesian (x, y, z) -> sx := !sx +. x; sy := !sy +. y; sz := !sz +. z
  | _ -> assert false) (Array.map to_cartesian points);
  Cartesian (!sx *. factor, !sy *. factor, !sz *. factor)

let rec plus p p' = match p, p' with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    Cartesian (x +. x', y +. y', z +. z')
  | _ -> plus (to_cartesian p) (to_cartesian p')

let rec times s = function
  | Cartesian (x, y, z) -> Cartesian (s *. x, s *. y, s *. z)
  | (Spherical _) as p -> times s (to_cartesian p)

let rec cross p p' = match p, p' with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    Cartesian (y *. z' -. z *. y', z *. x' -. x *. z', x *. y' -. y *. x')
  | _ -> cross (to_cartesian p) (to_cartesian p')

let magnitude = function
  | Cartesian (x, y, z) -> sqrt (x *. x +. y *. y +. z *. z)
  | Spherical _ -> 1.0

let norm = function
  | (Cartesian _) as p -> times (1.0 /. magnitude p) p
  | (Spherical _) as p -> p

let midpoint p p' =
  (* Midpoint of arc from p to p'
   *)
  norm (times 0.5 (plus p p'))

let incenter p1 p2 p3 =
  (* Calculate spherical incenter by intersecting angle bisectors.
   *
   * a12 = aux point at fixed dist along p1->p2
   * a13 = aux point at fixed dist along p1->p3
   * m1  = midpoint of a12->a13
   * p1->m1 bisects angle at p1
   *
   * Similarly a21, a23, m2, p2->m2 are used to bisect angle at p2.
   *
   * Intersect great circles of p1->m1 and p2->m2 to get antipodal
   * incenter candidates.  Pick the one that's closest to the centroid
   * of p1, p2, p3.
   *)
  let a12 = norm (plus p1 (norm (cross (cross p1 p2) p1))) in
  let a13 = norm (plus p1 (norm (cross (cross p1 p3) p1))) in
  let m1 = midpoint a12 a13 in
  let a21 = norm (plus p2 (norm (cross (cross p2 p1) p2))) in
  let a23 = norm (plus p2 (norm (cross (cross p2 p3) p2))) in
  let m2 = midpoint a21 a23 in
  let cand1 = norm (cross (cross p1 m1) (cross p2 m2)) in
  let cand2 = times (-1.0) cand1 in
  let dc = distance (center [| p1; p2; p3 |]) in
  if dc cand1 < dc cand2 then cand1 else cand2
