(* File: mirror_basics.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

open Printf
open OUnit
open Common

module P = Point
module M = Mirror

let genpoint () =
  let theta = Random.float pi and phi = Random.float (2. *. pi) in
  P.Spherical (theta, phi)

let genmirror () =
  let p1 = genpoint () and p2 = genpoint () in
  p1, p2, M.make_points p1 p2

let genimages m =
  let p = genpoint () in
  p, M.image m p

let origin = P.Cartesian (0., 0., 0.)

let tests = "Mirror basic tests" >::: [
  "Random spherical images always on the unit sphere" >:: (fun () ->
    Random.self_init ();
    for i = 0 to 10000 do
      let _, _, m = genmirror () in
      let p, p' = genimages m in
      sprintf "(%i) %s not on the unit sphere" i (P.to_string p') @?
        (P.distance origin p' =~ 1.)
    done
  );

  "Mirror made from 2 points contains those points and origin" >:: (fun () ->
    Random.self_init ();
    for i = 0 to 10000 do
      let p1, p2, m = genmirror () in
      "Origin not on the mirror" @? (M.distance m origin =~ 0.);
      "P1 not on the mirror" @? (M.distance m p1 =~ 0.);
      "P2 not on the mirror" @? (M.distance m p2 =~ 0.)
    done
  );

  "Distance from mirror to normal vector is 1" >:: (fun () ->
    Random.self_init ();
    for i = 0 to 10000 do
      let _, _, m = genmirror () in
      let d = M.distance m (P.Cartesian m) in
      sprintf "(%i) Distance is not 1 (is %f)" i d @? (d =~ 1.)
    done
  );

  "Mirror images at the same distance from the mirror" >:: (fun () ->
    Random.self_init ();
    for i = 0 to 10000 do
      let _, _, m = genmirror () in
      let p, p' = genimages m in
      let d1 = M.distance m p and d2 = M.distance m p' in
      sprintf "(%i) %.10f <> %.10f" i d1 d2 @? (cmp_float ~epsilon:1E-5 d1 d2)
    done
  );

  "image <-> original = image <-> mirror + mirror <-> original" >:: (fun () ->
    Random.self_init ();
    for i = 0 to 10000 do
      let _, _, m = genmirror () in
      let p, p' = genimages m in
      let d = M.distance m p and d' = M.distance m p' in
      let total = P.distance p p' in
      sprintf "(%i) %f + %f <> %f" i d d' total @? (d +. d' =~ total)
    done
  )
]
